/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp;

import com.google.common.base.Preconditions;

import cz.nocach.swingmvp.autoinitiater.MvpAutoInitiater;
import cz.nocach.swingmvp.model.Model;
import cz.nocach.swingmvp.model.ModelListenerValidator;
import cz.nocach.swingmvp.presenter.BasePresenter;
import cz.nocach.swingmvp.presenter.Presenter;
import cz.nocach.swingmvp.view.Change;
/**
 * main class for initializing model-view-presenter relationship.
 * @author __nocach
 *
 */
public class Mvp {
	/**
	 * Initializes connection between three main components.
	 * 1.Method ensures, that if {@code view} implements any of the model-listener's interfaces
	 * 	{@link cz.nocach.swingmvp.model.listener.Listener} then view will be added\set as listener
	 * of the model
	 * //TODO: write about defaultPropertyChange listeners (not implemented yet)
	 * 2.Establishes auto-delegation of view events to presenter of components with {@link Change}
	 * //TODO: document custom-methods of view and presenter in init phase (not implemented yet)
	 * 3.Each views' method, that implements interface of model-listener will be called with
	 * 	value that model have on the moment of this call. E.g. method {@code onNameChange(String name)}
	 *  is implements by {@code view} then it will be called with {@code model.getName()} value 
	 * //TODO: write about external implementation of model-listener (not implemented yet)
	 * //TODO: default-propertyChangeSupport does not work this way
	 * @param presenter not null presenter
	 * @param view not null view
	 * @param model not null model
	 * @param <V> type of view
	 * @param <M> type of model
	 * @throws NullPointerException if any parameter is null
	 * @throws MvpException if some validation error occurred while performing initiating mvp connection
	 */
	public static <V, M extends Model> void start(M model, V view,  BasePresenter<V, M> presenter){
		Preconditions.checkNotNull(presenter);
		Preconditions.checkNotNull(view);
		Preconditions.checkNotNull(model);
		new ModelListenerValidator(model).validate();
		//TODO: add view and presenter validation?
		MvpAutoInitiater mvpAutoInitiater = new MvpAutoInitiater(model, view);
		//before any manipulation make sure view is set as model listener
		mvpAutoInitiater.setViewAsModelListener();
		//now make sure view delegates any event to presenter
		mvpAutoInitiater.initViewPresenterAutoEventsDelegation(view, presenter);
		//now init view, every event in view will auto-init method on presenter
		mvpAutoInitiater.initViewFromModel();
	}
	/**
	 * works like {@link Mvp#start(Model, Object, BasePresenter)} but in addition in creates
	 * Presenter for view using {@link Presenter}.
	 * @return created presenter
	 * @throws MvpException if view has annotation @Presenter and {@code type} has 
	 * 		subclass of BasePresenter having no constructor with 2 params
	 * 		{@code V} and {@code M}
	 */
	public static <V,M extends Model> BasePresenter<V,M> start(M model, V view){
		Preconditions.checkNotNull(view);
		Preconditions.checkNotNull(model);
		MvpAutoInitiater mvpAutoInitiater = new MvpAutoInitiater(model, view);
		BasePresenter<V, M> presenter = mvpAutoInitiater.createPresenterForView();
		start(model, view, presenter);
		return presenter;
	}
}
