/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.autoinitiater;

import java.util.Collection;
import java.util.Set;

import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.model.Model;
import cz.nocach.swingmvp.model.ModelListenerExtractor;
import cz.nocach.swingmvp.util.CollectionsUtil;
import cz.nocach.swingmvp.util.FieldWithValue;
/**
 * 
 * @author __nocach
 *
 * @param <M> Model
 * @param <V> View
 */
public class ModelViewAutoSetter  <M extends Model, V> implements ViewAsModelSetter<M, V> {
	private final M model;
	private final V view;
	private ModelListenerExtractor listenerExtractor;
	public ModelViewAutoSetter(M model, V view){
		this.model = model;
		this.view = view;
		this.listenerExtractor = new ModelListenerExtractor();
	}
	/* (non-Javadoc)
	 * @see cz.nocach.swingmvp.autoinitiater.ViewAsModelSetter#setViewAsModelListener()
	 */
	@Override
	public void setViewAsModelListener() {
		ModelListenerExtractor listenerExtractor = new ModelListenerExtractor();
		Set<Class<?>> viewCanImplementThisClasses = listenerExtractor.getPossibleListenerClasses(model);
		boolean viewMustBeSetAsListener 
			= checkIfViewImplementsAtLeastOneModelListener(viewCanImplementThisClasses);
		if (viewMustBeSetAsListener){
			setViewAsListenerForEachImplementingInterface();
		}
		else{
			assertModelHasAtLeastOneListenerTypeInstance(listenerExtractor
					, viewCanImplementThisClasses);
		}
	}
	
	private boolean checkIfViewImplementsAtLeastOneModelListener(Set<Class<?>> viewCanImplementThisClasses) {
		for(Class<?> each : viewCanImplementThisClasses){
			for(Class<?> eachViewImplementingInterface : view.getClass().getInterfaces()){
				if (eachViewImplementingInterface.isAssignableFrom(each)) return true;
			}
		}
		return false;
	}
	
	private void setViewAsListenerForEachImplementingInterface() {
		for(Class<?> possibleListenerType : view.getClass().getInterfaces()){
			if (!isModelListenerClass(possibleListenerType)) continue;
			Set<FieldWithValue> modelListenerFields = 
					listenerExtractor.getListenerFieldsWithValuesFor(model, possibleListenerType);
			if (isViewAlreadyAssignedAsListenerOnModel(possibleListenerType, modelListenerFields)) continue;
			setViewAsModelListener(possibleListenerType, modelListenerFields);
		}
	}
	
	private boolean isModelListenerClass(Class<?> typeToTest) {
		for (Class<?> listenerType : listenerExtractor.getPossibleListenerClasses(model)){
			if (listenerType.isAssignableFrom(typeToTest)) return true;
		}
		return false;
	}

	private void setViewAsModelListener(Class<?> implementedListenerType, Set<FieldWithValue> modelListenerFields) {
		setViewAsFirstAvailableListener(implementedListenerType, modelListenerFields);
	}
	
	private void assertModelHasAtLeastOneListenerTypeInstance(
			ModelListenerExtractor listenerExtractor, Set<Class<?>> listenerTypes) {
		for(Class<?> eachListenerType : listenerTypes){
			boolean typeHadInstance = false;
			Set<FieldWithValue> currentListenersFields 
				= listenerExtractor.getListenerFieldsWithValuesFor(model, eachListenerType);
			for (FieldWithValue eachField : currentListenersFields){
				if (fieldIsCollectionOfListeners(eachField)){
					Collection<?> listeners = (Collection<?>)eachField.value;
					if (CollectionsUtil.hasAtLeastOneNotNullElement(listeners)) {
						typeHadInstance = true;
						break;
					}
				}
				else{
					if (eachField.value != null) {
						typeHadInstance = true;
						break;
					}
				}
			}
			if (typeHadInstance) return;
		}
		throw new MvpException("Model must have atleast one listener assigned when " +
				"view does not implement any of the model listeners");
	}
	
	@SuppressWarnings("rawtypes")
	private boolean isViewAlreadyAssignedAsListenerOnModel(Class<?> implementedListenerType, 
			Set<FieldWithValue> modelListenerFields) {
		for(FieldWithValue each : modelListenerFields){
			if (implementedListenerType.isAssignableFrom(each.field.getType())
					&& each.value == view){
				//view is already assigned
				return true;
			}
			if (fieldIsCollection(each)){
				if (each.value == null){
					throw new MvpException("collection of listeners with annotation @Listener " +
							"must not be null");
				}
				else{
					if (((Collection)each.value).contains(view)) return true;
				}
			}
		}
		return false;
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void setViewAsFirstAvailableListener(Class<?> implementedListenerType,
			Set<FieldWithValue> listenerProperties) {
		for(FieldWithValue each : listenerProperties){
			if (fieldIsNotAssignedListenerField(implementedListenerType, each)){
				each.field.setAccessible(true);
				try {
					each.field.set(model, view);
				} catch (IllegalArgumentException e) {
					// TODO Handle
					throw new RuntimeException(e);
				} catch (IllegalAccessException e) {
					// TODO Handle
					throw new RuntimeException(e);
				}
				return;
			}
			if (fieldIsCollectionOfListeners(each)){
				if (each.value == null){
					throw new MvpException("collection of listeners with annotation @Listener " +
							"must not be null");
				}
				else{
					((Collection)each.value).add(view);
					return;
				}
			}
		}
		throw new MvpException("View coudn't be assigned as listener, because model already has" +
				" another listener instance assigned");
	}
	
	private boolean fieldIsCollectionOfListeners(FieldWithValue fieldWIthValue) {
		return fieldIsCollection(fieldWIthValue);
	}
	
	private boolean fieldIsCollection(FieldWithValue each) {
		return Collection.class.isAssignableFrom(each.field.getType());
	}

	private boolean fieldIsNotAssignedListenerField(
			Class<?> implementedListenerType, FieldWithValue each) {
		return implementedListenerType.isAssignableFrom(each.field.getType())
				&& each.value == null;
	}
}
