/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.autoinitiater;

import static com.google.common.base.Preconditions.*;
import static com.google.common.base.Preconditions.checkNotNull;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Set;

import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.model.Model;
import cz.nocach.swingmvp.model.ModelListenerExtractor;
import cz.nocach.swingmvp.model.ModelListenerInvoker;
import cz.nocach.swingmvp.presenter.BasePresenter;
import cz.nocach.swingmvp.presenter.Presenter;
import cz.nocach.swingmvp.util.Annotations;
import cz.nocach.swingmvp.util.CollectionsUtil;
import cz.nocach.swingmvp.util.FieldWithValue;
import cz.nocach.swingmvp.view.Binders;
import cz.nocach.swingmvp.view.Change;
import cz.nocach.swingmvp.view.WidgetBinder;
/**
 * class performing real-initiating logic of Model-View-Presenter connection
 * @author __nocach
 *
 */
public class MvpAutoInitiater <M extends Model, V> {
	
	private final M model;
	private final V view;
	private ViewAsModelSetter<M, V> modelViewAutoSetter;

	public MvpAutoInitiater(M model, V view){
		checkNotNull(view);
		checkNotNull(model);
		
		this.model = model;
		this.view = view;
		modelViewAutoSetter = createModelViewAutoSetter(model, view);
	}

	protected ViewAsModelSetter<M, V> createModelViewAutoSetter(M model, V view) {
		return new ModelViewAutoSetter<M, V>(model, view);
	}

	/**
	 * invokes methods of {@code view} that are implementations of any listener-interfaces of the model with
	 * properties from the model. If view does not implement any listener-interfaces, then nothing will happen
	 */
	public void initViewFromModel() {
		new ModelListenerInvoker(model).invokeCustomListenerEventForEachModelPropertyValue(view);
	}

	/**
	 * walks through components of the {@code view} and looks for fields with {@link Change}
	 * annotation. For each such annotation framework will create delegation of events from those components.
	 * E.g. if field is JTextField then any text change event will be delegated to presenter
	 * @param presenter not null presenter
	 * @param view not null view of the presenter
	 * @throws NullPointerException if {@code view} or {@code presenter} are null
	 * @throws MvpException if component with {@link Change} has no approperiate WidgetBinder
	 */
	@SuppressWarnings("unchecked")
	public <V> void initViewPresenterAutoEventsDelegation(V view, BasePresenter<V, ?> presenter) {
		checkNotNull(presenter);
		checkNotNull(view);
		for(Field each : view.getClass().getDeclaredFields()){
			try {
				each.setAccessible(true);
				//TODO: make extensiable for annotation types
				Change changeAnnotation = Annotations.getDeclaredAnnotationIfPresent(each, Change.class);
				if (changeAnnotation != null) {
					@SuppressWarnings("rawtypes")
					WidgetBinder binderForFieldComponent = Binders.forComponent(each.getType());
					Object widget = each.get(view);
					String methodOnPresenter = changeAnnotation.value();
					binderForFieldComponent.bindTo(widget, presenter, methodOnPresenter);
				}
			} catch (IllegalAccessException e) {
				// TODO: Handle
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * if view implements listeners of the model (will find out listeners by observing
	 * properties with annotation @Listener on model) this will automatically assign to such
	 * listeners passed view. If view is already assigned as listener, then nothing will happen
	 * @throws MvpException if some validation error while performing this auto set has occurred
	 */
	public void setViewAsModelListener() {
		modelViewAutoSetter.setViewAsModelListener();
	}


	/**
	 * 
	 * @throws MvpException if view has annotation @Presenter and {@code type} has 
	 * 		subclass of BasePresenter having no constructor with 2 params
	 * 		{@code V} and {@code M}
	 * @throws NullPointerException if view is null
	 */
	public BasePresenter<V,M> createPresenterForView(){
		checkNotNull(view);
		Presenter presenterOnView = Annotations.getDeclaredAnnotationIfPresent(
				view.getClass(), Presenter.class);
		if (presenterOnView == null) return new BasePresenter<V, M>(view, model);
		for (Constructor each : presenterOnView.type().getConstructors()){
			if (each.getParameterTypes().length != 2) continue;
			if (each.getParameterTypes()[0].isInstance(view)
					&& each.getParameterTypes()[1].isInstance(model)){
				try {
					return (BasePresenter<V, M>) each.newInstance(view, model);
				} catch (IllegalArgumentException e) {
					// TODO Handle
					throw new RuntimeException(e);
				} catch (InstantiationException e) {
					// TODO Handle
					throw new RuntimeException(e);
				} catch (IllegalAccessException e) {
					// TODO Handle
					throw new RuntimeException(e);
				} catch (InvocationTargetException e) {
					// TODO Handle
					throw new RuntimeException(e);
				}
			}
		}
		throw new MvpException(presenterOnView.type() + " must have constructor with two params" +
				" (view,model)");
	}

	public ViewAsModelSetter<M, V> getModelViewAutoSetter() {
		return modelViewAutoSetter;
	}
	public void setModelViewAutoSetter(ViewAsModelSetter<M, V> modelViewAutoSetter) {
		checkNotNull(modelViewAutoSetter);
		this.modelViewAutoSetter = modelViewAutoSetter;
	}
}
