/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.model;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.jgoodies.binding.beans.BeanAdapter;

import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.model.listener.CustomPropertyListenerInvokeInterface;
import cz.nocach.swingmvp.model.listener.PropertyListenerInvokeInterface;
import cz.nocach.swingmvp.model.listener.Listener;
import cz.nocach.swingmvp.model.listener.ListenerException;
import cz.nocach.swingmvp.util.Annotations;

/**
 * base class for PresentationModel
 * subclassing from this class is essential for auto-event firing and other things to work
 * @author __nocach
 *
 */
public class Model {
	protected PropertyChangeSupport changeSupport = new PropertyChangeSupport(this);
	ModelListenerInvoker listenerInvoker = new ModelListenerInvoker(this);
	boolean defaultChangeSupportReplacedByCustom = true;
	Map<Object, PropertyListenerInvokeInterface> listenersWithInvokeInterfaces
		= new HashMap<Object, PropertyListenerInvokeInterface>();
	/**
	 * if class has any custom listener, then they all be stored to this collection
	 */
	Set<Object> customListeners = new HashSet<Object>();
	/**
	 * holds cached Fields that had {@link Listener} annotation on them
	 */
	Set<Field> fieldsWithListenerAnnotation;
	/**
	 * 
	 * @param listener listener to be added to model
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener){
		if (defaultChangeSupportReplacedByCustom){
			throw new ListenerException("default change support was replaced by custom listener support");
		}
		changeSupport.addPropertyChangeListener(listener);
	}
	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener){
		if (defaultChangeSupportReplacedByCustom){
			throw new ListenerException("default change support was replaced by custom listener support");
		}
		changeSupport.addPropertyChangeListener(propertyName, listener);
	}
	/**
	 * 
	 * @param listener listener to be added to model
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener){
		if (defaultChangeSupportReplacedByCustom){
			throw new ListenerException("default change support was replaced by custom listener support");
		}
		changeSupport.removePropertyChangeListener(listener);
	}
	/**
	 * 
	 * @param listener listener to be added to model
	 */
	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener){
		if (defaultChangeSupportReplacedByCustom){
			throw new ListenerException("default change support was replaced by custom listener support");
		}
		changeSupport.removePropertyChangeListener(propertyName, listener);
	}
	/**
	 * goes through all model-fields and searches for fields with @Listener annotation
	 * instances from those fields (both single-value and collections) will be added to {@code customListeners}
	 */
	void initCustomListeners(){
		Model model = this;
		model.customListeners.clear();
		for (Field eachField : getFieldsWithListenerAnnotation()){
			try {
				eachField.setAccessible(true);
				//check if this field is collection of Listeners
				if (Collection.class.isAssignableFrom(eachField.getType())){
					Collection<?> customListeners = (Collection<?>)eachField.get(model);
					if (customListeners == null) continue;
					model.customListeners.addAll(customListeners);
				}
				else{
					model.customListeners.add(eachField.get(model));
				}
			} catch (IllegalArgumentException e) {
				//TODO: handle
				throw new MvpException(e);
			} catch (IllegalAccessException e) {
				//TODO: handle
				throw new MvpException(e);
			}
		}
		//remove null listeners if any
		model.customListeners.remove(null);
	}
	
	private Set<Field> getFieldsWithListenerAnnotation(){
		if (fieldsWithListenerAnnotation != null) return fieldsWithListenerAnnotation;
		fieldsWithListenerAnnotation = new HashSet<Field>();
		for (Field eachField : getClass().getDeclaredFields()){
			for(Annotation eachAnnotation : eachField.getDeclaredAnnotations()){
				if (eachAnnotation instanceof Listener){
					fieldsWithListenerAnnotation.add(eachField);
				}
			}
		}
		return fieldsWithListenerAnnotation;
	}
	
	/**
	 * sets custom property on model and notifies all listeners about this change
	 * @param propertyName name of the property
	 * @param property value to be set
	 */
	public void setProperty(String propertyName, Object property){
		BeanAdapter<Object> modelBeanAdapter = new BeanAdapter<Object>(this);
		Object oldValue = modelBeanAdapter.getValue(propertyName);
		modelBeanAdapter.setValue(propertyName, property);
		listenerInvoker.processPropertyChange(propertyName, oldValue, property);
	}
	void initListenerAndTheirInvokeInterfaces() {
		listenersWithInvokeInterfaces.clear();
		if (!defaultChangeSupportReplacedByCustom){
			for(PropertyChangeListener each : changeSupport.getPropertyChangeListeners()){
				listenersWithInvokeInterfaces.put(each, new BeansPropertyChangeListenerInvokeInterface());
			}
		}
		for (Field eachField : getFieldsWithListenerAnnotation()){
			putListenerWithInvokeInterfaceFromField(eachField);
		}
		//null instances must be removed
		listenersWithInvokeInterfaces.remove(null);
	}
	private void putListenerWithInvokeInterfaceFromField(Field eachField) {
		try {
			eachField.setAccessible(true);
			//check if this field is collection of Listeners
			if (Collection.class.isAssignableFrom(eachField.getType())){
				Collection<?> customListeners = (Collection<?>)eachField.get(this);
				if (customListeners == null)
					return;
				for(Object eachCustomListener : customListeners){
					Listener listenerAnnotation = Annotations.getDeclaredAnnotationIfPresent(eachField, Listener.class);
					CustomPropertyListenerInvokeInterface listenerInvokeInterface
						= new CustomPropertyListenerInvokeInterface(listenerAnnotation.type());
					listenersWithInvokeInterfaces.put(eachCustomListener, listenerInvokeInterface);
				}
			}
			else{
				CustomPropertyListenerInvokeInterface listenerInvokeInterface
					= new CustomPropertyListenerInvokeInterface(eachField.getType());
				listenersWithInvokeInterfaces.put(eachField.get(this), listenerInvokeInterface);
			}
		} catch (IllegalArgumentException e) {
			//TODO: handle
			throw new MvpException(e);
		} catch (IllegalAccessException e) {
			//TODO: handle
			throw new MvpException(e);
		}
	}
}
