/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.model;

import static com.google.common.base.Preconditions.checkNotNull;


import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.model.listener.Listener;
import cz.nocach.swingmvp.model.listener.ListenerHelper;
import static cz.nocach.swingmvp.model.listener.ListenerHelper.*;
import cz.nocach.swingmvp.util.Annotations;
import cz.nocach.swingmvp.util.FieldWithValue;
/**
 * class for extracting Listener interfaces and fields of that interfaces 
 * from model
 * @author __nocach
 *
 */
public class ModelListenerExtractor {

	/**
	 * 
	 * @param model not null model with valid listener fields, see {@link ModelListenerValidator}
	 * @return not null set with classes of interfaces of listeners that model can have
	 * @throws NullPointerException if model is null
	 */
	public Set<Class<?>> getPossibleListenerClasses(Model model) {
		checkNotNull(model);
		Set<Class<?>> result = new HashSet<Class<?>>();
		for(Field eachField : model.getClass().getDeclaredFields()){
			if (Collection.class.isAssignableFrom(eachField.getType())){
				Listener listenerDescriptor = Annotations.getDeclaredAnnotationIfPresent(eachField, Listener.class);
				if (listenerDescriptor == null) continue; 
				result.add(listenerDescriptor.type());
			}
			else{
				if (Annotations.getDeclaredAnnotationIfPresent(eachField, Listener.class) != null){
					result.add(eachField.getType());
				}
			}
		}
		return result;
	}

	/**
	 * 
	 * @param model not null model
	 * @param listenerType not null type of the listener, which {@link FieldWithValue} must be returned
	 * @return set of FieldWithValue. Each field can have null property 
	 *		listener instance or collection of listeners.
	 * @throws NullPointerException if {@code model} or {@code type} are null
	 */
	public <T> Set<FieldWithValue> getListenerFieldsWithValuesFor(Model model, Class<T> listenerType){
		checkNotNull(model);
		checkNotNull(listenerType);
		Set<FieldWithValue> result = new HashSet<FieldWithValue>();
		for (Field eachField : model.getClass().getDeclaredFields()){
			if (!isListenerField(eachField)) continue;
			Class<?> listenerTypeOfTheField = extractListenerTypeFromField(eachField);
			if (!listenerTypeOfTheField.isAssignableFrom(listenerType)) continue;
			eachField.setAccessible(true);
			try {
				result.add(new FieldWithValue(eachField,eachField.get(model)));
			} catch (IllegalArgumentException e) {
				// TODO Handle
				throw new MvpException(e);
			} catch (IllegalAccessException e) {
				// TODO Handle
				throw new MvpException(e);
			}
		}
		return result;
	}
	
	private boolean isListenerField(Field field){
		Listener listenerDescriptor = Annotations.getDeclaredAnnotationIfPresent(field, Listener.class);
		return listenerDescriptor != null;
	}

	private Class<?> extractListenerTypeFromField(Field field) {
		Listener listenerDescriptor = Annotations.getDeclaredAnnotationIfPresent(field, Listener.class);
		if (doesListenerDescriptorHasNoTypeSpecified(listenerDescriptor)) return field.getType();
		return listenerDescriptor.type();
	}

	private <T> boolean isFieldOfListenerType(Field field, Class<T> listenerType) {
		return field.getType().isAssignableFrom(listenerType);
	}
}
