/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import cz.nocach.swingmvp.model.listener.Listener;

/**
 * aspect for supporting automatic firing of property change events to defaultPropertyChangeSupport
 * or any custom listeners of the model extending {@link Model}
 * @author __nocach
 *
 */
public aspect ModelListenerSupportAspect {
	pointcut modelInit():execution(Model.new(..));
	
	/**
	 * advice determinating if model should use only custom-listeners
	 * @param model
	 */
	after(Model model) : modelInit() && this(model){
		setDefaultChangeSupportEnabled(model);
	}
	
	private void setDefaultChangeSupportEnabled(Model model){
		model.defaultChangeSupportReplacedByCustom = false;
		for(Field eachField : model.getClass().getDeclaredFields()){
			for(Annotation eachAnnotation : eachField.getDeclaredAnnotations()){
				if (eachAnnotation instanceof Listener){
					model.defaultChangeSupportReplacedByCustom = true;
					return;
				}
			}
		}
	}
	
	pointcut setter(cz.nocach.swingmvp.model.Model model):
		target(model)
		&& call(* Model+.set*(*));

	/**
	 * fires event of the simple set methods
	 * @param model model on which set method was called
	 */
	void around(cz.nocach.swingmvp.model.Model model) : setter(model) {
		try {
			String propertyMethodName = thisJoinPoint.getSignature().getName().substring("set".length());
			String propertyName = propertyMethodName.substring(0, 1).toLowerCase()+ propertyMethodName.substring(1);
			Method propertyGetter = model.getClass().getMethod("get" + propertyMethodName);
			Object beforeSetPropertyValue = propertyGetter.invoke(model);
			proceed(model);
			Object afterSetPropertyValue = propertyGetter.invoke(model);
			
			model.listenerInvoker.processPropertyChange(propertyName, beforeSetPropertyValue, afterSetPropertyValue);
		} catch (SecurityException e) {
			//TODO: handle
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			//TODO: handle
			throw new RuntimeException(e);
		} catch (IllegalArgumentException e) {
			//TODO: handle
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			//TODO: handle
			throw new RuntimeException(e);
		} catch (InvocationTargetException e) {
			//TODO: handle
			throw new RuntimeException(e);
		}
	}
	pointcut collectionOperation(Model model):
		target(model)
		&& (call(* Model+.add*(*)) || call(* Model+.remove*(*)));
	
	/**
	 * fires collection-related event after collection event operation 
	 * @param model model on which set method was called
	 */
	after(cz.nocach.swingmvp.model.Model model) : collectionOperation(model){
		boolean isAddMethod = thisJoinPoint.getSignature().getName().startsWith("add");
		String propertyMethodName = isAddMethod 
				? thisJoinPoint.getSignature().getName().substring("add".length())
				: thisJoinPoint.getSignature().getName().substring("remove".length());
		String propertyName = Character.toLowerCase(propertyMethodName.charAt(0)) + 
				propertyMethodName.substring(1);
		//collection operations will always have 1 arg, add({arg}) and remove({arg})
		Object property = thisJoinPoint.getArgs()[0];
		if (isAddMethod){
			model.listenerInvoker.firePropertyAdd(propertyName, property);
		}
		else{
			model.listenerInvoker.firePropertyRemoved(propertyName, property);
		}
	}

}
