/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.model;

import static com.google.common.base.Preconditions.checkNotNull;

import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Set;

import cz.nocach.swingmvp.model.listener.Add;
import cz.nocach.swingmvp.model.listener.CollectionFromListenerNotExistException;
import cz.nocach.swingmvp.model.listener.Listener;
import cz.nocach.swingmvp.model.listener.ListenerException;
import cz.nocach.swingmvp.model.listener.ListenerHelper;
import cz.nocach.swingmvp.model.listener.ListenerTypeNotSpecified;
import cz.nocach.swingmvp.model.listener.Remove;
import cz.nocach.swingmvp.util.Annotations;
import cz.nocach.swingmvp.util.Fields;
/**
 * 
 * @author __nocach
 *
 */
public class ModelListenerValidator {
	private final Model model;
	public ModelListenerValidator(Model model) {
		this.model = checkNotNull(model);
	}
	/**
	 * make sure that {@link Add} and {@link Remove} have valid values on listener interfaces
	 * of the model
	 * @throws ListenerException
	 */
	public void checkListenerHasValidCollectionPropertyAnnotations() throws ListenerException{
		ModelListenerExtractor listenerExtractor = new ModelListenerExtractor();
		Set<Class<?>> listenerClasses = listenerExtractor.getPossibleListenerClasses(model);
		for (Class<?> eachListenerClass : listenerClasses){
			assertAddAnnotationHasValidCollectionName(eachListenerClass);
			assertRemoveAnnotationHasValidCollectionName(eachListenerClass);
		}
	}
	private void assertRemoveAnnotationHasValidCollectionName(
			Class<?> eachListenerClass) {
		for(Remove eachRemoveAnnotation : Annotations.getAnnotationsOnDeclaredMethods(eachListenerClass, Remove.class)){
			try {
				Fields.getDeclaredFieldFromHierarchy(model.getClass(), eachRemoveAnnotation.collection());
			} catch (SecurityException e) {
				throw new CollectionFromListenerNotExistException(eachRemoveAnnotation.collection()
						, eachRemoveAnnotation, e);
			} catch (NoSuchFieldException e) {
				throw new CollectionFromListenerNotExistException(eachRemoveAnnotation.collection()
						, eachRemoveAnnotation, e);
			}
		}
	}
	private void assertAddAnnotationHasValidCollectionName(
			Class<?> eachListenerClass) {
		for(Add eachAddAnnotation : Annotations.getAnnotationsOnDeclaredMethods(eachListenerClass, Add.class)){
			try {
				Fields.getDeclaredFieldFromHierarchy(model.getClass(), eachAddAnnotation.collection());
			} catch (SecurityException e) {
				throw new CollectionFromListenerNotExistException(eachAddAnnotation.collection()
						, eachAddAnnotation, e);
			} catch (NoSuchFieldException e) {
				throw new CollectionFromListenerNotExistException(eachAddAnnotation.collection()
						, eachAddAnnotation, e);
			}
		}
	}
	
	/**
	 * walks through @Listener annotations on fields subtypes of Collection and checks
	 * if the have {@link Listener#type()} specified to make sure erasure will not stand in
	 * our way
	 */
	public void checkCollectionOfListenersHasTypeSpecified() {
		for (Field each : model.getClass().getDeclaredFields()){
			if (Collection.class.isAssignableFrom(each.getType())){
				Listener listenerAnnotation = Annotations.getDeclaredAnnotationIfPresent(each, Listener.class);
				if (listenerAnnotation == null) continue;
				if (ListenerHelper.doesListenerDescriptorHasNoTypeSpecified(listenerAnnotation)){
					throw new ListenerTypeNotSpecified();
				}
			}
		}
	}
	
	/**
	 * perform all validation
	 */
	public void validate(){
		checkCollectionOfListenersHasTypeSpecified();
		checkListenerHasValidCollectionPropertyAnnotations();
	}

}
