/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.util;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.Set;

import org.testng.collections.Sets;

import com.google.common.base.Function;
import com.google.common.base.Preconditions;

import cz.nocach.swingmvp.testobj.ValueAnnotation;
/**
 * helper class for extracting and manipulating with annotations
 * @author __nocach
 *
 */
public class Annotations {
	/**
	 * will go through all interface of each passed objects and will search for methods on that objects
	 * which interfaces will have annotation {@code annotation} and will satisfy optional 
	 * filter {@code acceptAnnotation}. For each such object and method instance of {@link ObjectsMethodWithAnnotation} will
	 * be created and passed to {@code applyOnObjectMethod}
	 * @param objects not null collection of objects that can have methods with annotation that is 
	 * 			instanceof {@code annotation}
	 * @param annotation not null class of annotation that is anticipated on methods
	 * @param acceptAnnotation optional function that determinates if {@code applyOnObjectMethod} must be called
	 * 			for method with annotation of type {@code annotation}
	 * @param applyOnObjectMethod not null function that will receive pair of Object and it's method 
	 * 			that had annotation {@code annotation} and passed {@code acceptAnnotation} filter
	 */
	public static <T extends Annotation> void applyForObjectsMethodWithAnnotation(Collection<Object> objects, Class<T> annotation,
			Function<T, Boolean> acceptAnnotation, Function<ObjectsMethodWithAnnotation<T>, Void> applyOnObjectMethod){
		Preconditions.checkNotNull(objects);
		Preconditions.checkNotNull(annotation);
		Preconditions.checkNotNull(applyOnObjectMethod);
		for(Object eachObject : objects){
			//@Property annotation resides on interfaces.
			//model.customListeners has realizations that have Overriden methods without annotations on interfaces.
			//Check all methods of interfaces for @Property annotation to find all possible methods with
			//@Property annotation
			for(Class<?> eachListenerInterface : eachObject.getClass().getInterfaces()){
				for(Method eachMethod : eachListenerInterface.getDeclaredMethods()){
					for(Annotation eachAnnotation : eachMethod.getDeclaredAnnotations()){
						if (eachAnnotation.annotationType().isAssignableFrom(annotation)){
							if (acceptAnnotation == null || acceptAnnotation.apply(annotation.cast(eachAnnotation))){
								try {
									Method methodOnObject = eachObject.getClass().getDeclaredMethod(
											eachMethod.getName() , eachMethod.getParameterTypes());
									methodOnObject.setAccessible(true);
									applyOnObjectMethod.apply(
											new ObjectsMethodWithAnnotation<T>(eachObject, methodOnObject,
													annotation.cast(eachAnnotation)));
								} catch (SecurityException shouldNotHappen) {
									throw new RuntimeException(shouldNotHappen);
								} catch (NoSuchMethodException shouldNotHappen) {
									throw new RuntimeException(shouldNotHappen);
								}
							}
							//we now have method of interface, so get Method on the Listener itself
							break;
						}
					}
				}
			}
		}
	}
	/**
	 * will go through all interface of each passed objects and will search for methods on that objects
	 * which interfaces will have annotation {@code annotation} and will satisfy optional 
	 * filter {@code acceptAnnotation}. For each such object and method instance of {@link ObjectsMethodWithAnnotation} will
	 * be created and passed to {@code applyOnObjectMethod}
	 * @param object not null object that can have methods with annotation that is 
	 * 			instanceof {@code annotation}
	 * @param annotation not null class of annotation that is anticipated on methods
	 * @param acceptAnnotation optional function that determinates if {@code applyOnObjectMethod} must be called
	 * 			for method with annotation of type {@code annotation}
	 * @param applyOnObjectMethod not null function that will receive pair of Object and it's method 
	 * 			that had annotation {@code annotation} and passed {@code acceptAnnotation} filter
	 */
	public static <T extends Annotation> void applyForObjectMethodWithAnnotation(Object object, Class<T> annotation,
			Function<T, Boolean> acceptAnnotation, Function<ObjectsMethodWithAnnotation<T>, Void> applyOnObjectMethod){
		applyForObjectsMethodWithAnnotation(Collections.<Object>singleton(object),
				annotation, acceptAnnotation, applyOnObjectMethod);
	}
	
	/**
	 * @param field field to be inspected on declared annotations
	 * @param annotationType type of the annotation to search
	 * @return annotation of the {@code annotationType} or null if declared annotation is
	 * 	not present
	 */
	public static <T extends Annotation> T getDeclaredAnnotationIfPresent(Field field, Class<T> annotationType){
		for(Annotation each : field.getDeclaredAnnotations()){
			if (annotationType.isAssignableFrom(each.getClass())){
				return annotationType.cast(each);
			}
		}
		return null;
	}
	/**
	 * @param method method to be inspected on declared annotations
	 * @param annotationType type of the annotation to search
	 * @return annotation of the {@code annotationType} or null if declared annotation is
	 * 	not present
	 */
	public static <T extends Annotation> T getDeclaredAnnotationIfPresent(Method method, Class<T> annotationType){
		for(Annotation each : method.getDeclaredAnnotations()){
			if (annotationType.isAssignableFrom(each.getClass())){
				return annotationType.cast(each);
			}
		}
		return null;
	}
	/**
	 * @param clazz class to be inspected on declared annotations
	 * @param annotationType type of the annotation to search
	 * @return annotation of the {@code annotationType} or null if declared annotation is
	 * 	not present
	 */
	public static <T extends Annotation> T getDeclaredAnnotationIfPresent(Class<?> clazz, Class<T> annotationType){
		for(Annotation each : clazz.getDeclaredAnnotations()){
			if (annotationType.isAssignableFrom(each.getClass())){
				return annotationType.cast(each);
			}
		}
		return null;
	}
	/**
	 * goes through all declared methods of the {@code clazz} and searches for annotations with
	 * type {@code annotationType}  
	 * @param clazz not null clazz 
	 * @param annotationType not null annotationType
	 * @return not null {@link Collection} of found annotation
	 * @throws NullPointerException if {@code clazz} or {@code annotationType} are null
	 */
	public static <T> Collection<T> getAnnotationsOnDeclaredMethods(Class<?> clazz, Class<T> annotationType) {
		Preconditions.checkNotNull(clazz);
		Preconditions.checkNotNull(annotationType);
		Set<T> result = Sets.newHashSet();
		for(Method eachMethod: clazz.getDeclaredMethods()){
			for(Annotation eachAnnotation : eachMethod.getDeclaredAnnotations()){
				if (annotationType.isAssignableFrom(eachAnnotation.getClass())){
					result.add(annotationType.cast(eachAnnotation));
				}
			}
		}
		return result;
	}
	/**
	 * 
	 * @param clazz
	 * @param annotationType
	 * @return //TODO: write doc
	 */
	public static <T> Collection<T> getAnnotationsOnDeclaredFields(
			Class<?> clazz,
			Class<T> annotationType) {
		Preconditions.checkNotNull(clazz);
		Preconditions.checkNotNull(annotationType);
		Set<T> result = Sets.newHashSet();
		for(Field eachField: clazz.getDeclaredFields()){
			for(Annotation eachAnnotation : eachField.getDeclaredAnnotations()){
				if (annotationType.isAssignableFrom(eachAnnotation.getClass())){
					result.add(annotationType.cast(eachAnnotation));
				}
			}
		}
		return result;
	}
}
