/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.mvp;

import java.util.Collection;
import java.util.HashSet;

import org.jmock.Mockery;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import cz.nocach.swingmvp.BaseTest;
import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.autoinitiater.ModelViewAutoSetter;
import cz.nocach.swingmvp.autoinitiater.MvpAutoInitiater;
import cz.nocach.swingmvp.model.Model;
import cz.nocach.swingmvp.model.listener.Listener;
import cz.nocach.swingmvp.testmodels.Listener1;
import cz.nocach.swingmvp.testmodels.Listener2;
import cz.nocach.swingmvp.testmodels.ListenerParent;
import cz.nocach.swingmvp.testmodels.ModelWithMultipleListeners;
import cz.nocach.swingmvp.testmodels.ModelWithSingleListener;

public class ModelViewAutoSetterTest extends BaseTest{
	
	private Listener1 listener1ExternalInstance;
	private Listener2 listener2ExternalInstance1;
	private Listener2 listener2ExternalInstance2;

	@BeforeMethod
	public void setUp(){
		listener1ExternalInstance = mockery.mock(Listener1.class, "listener1ExternalInstance");
		listener2ExternalInstance1 = mockery.mock(Listener2.class, "listener2ExternalInstance1");
		listener2ExternalInstance2 = mockery.mock(Listener2.class, "listener2ExternalInstance2");
	}
	
	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	public void autoSettingOfViewAsModelListener(){
		ModelWithMultipleListeners model = new ModelWithMultipleListeners();
		Listener1 view = mockery.mock(Listener1.class);
		assertFalse(model.hasAssignedListener1Value(view));
		
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(model, view);
		autoSetter.setViewAsModelListener();
		
		assertTrue(model.hasAssignedListener1Value(view));
	}
	
	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	//TODO: add test for view implementing javabean's PropertyChangeListener
	public void allListenerInterfaces_ImplementedByView_MustBeSetOnModel(){
		final ModelWithMultipleListeners model = new ModelWithMultipleListeners();
		ViewImplementingListener1AndListener2 view = new ViewImplementingListener1AndListener2();
		addExternalListenersToModel(model);
		
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(model, view);
		autoSetter.setViewAsModelListener();

		assertTrue(model.hasAssignedListener1Value(view));
		assertTrue(model.hasAssignedListener2Value(view));
		//previously assigned were preserved
		assertAddedExternalListenersPreserved(model);
		
	}
	
	private void addExternalListenersToModel(
			final ModelWithMultipleListeners model) {
		model.listener1 = listener1ExternalInstance;
		model.listener2 = listener2ExternalInstance1;
		model.listenersListener2.add(listener2ExternalInstance2);
	}

	private void assertAddedExternalListenersPreserved(
			final ModelWithMultipleListeners model) {
		assertEquals(model.listener1, listener1ExternalInstance);
		assertEquals(model.listener2, listener2ExternalInstance1);
		assertEquals(model.listenersListener2.size(), 2);
	}

	
	@Test
	public void noAssignIfViewIsAlreadyAssigned(){
		ModelWithMultipleListeners modelWithListenersWithViewAlreadyAssigned
			= new ModelWithMultipleListeners();
		ViewImplementingListener1AndListener2 view = new ViewImplementingListener1AndListener2();
		modelWithListenersWithViewAlreadyAssigned.listener1 = view;
		modelWithListenersWithViewAlreadyAssigned.listenersListener2.add(view);
		
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(modelWithListenersWithViewAlreadyAssigned, view);
		autoSetter.setViewAsModelListener();
		
		assertNull(modelWithListenersWithViewAlreadyAssigned.listener1AnotherField);
		assertEquals(modelWithListenersWithViewAlreadyAssigned.listenersListener1.size(), 0);
		assertEquals(modelWithListenersWithViewAlreadyAssigned.listenersListener2.size(), 1);
		//TODO: Sometimes fails here, probably because of the order of walking through fields.
		//TODO: maybe write more specific tests
		assertNull(modelWithListenersWithViewAlreadyAssigned.listener2);
	}
	
	@Test
	public void noAutoAssignIfViewIsPresentedInCollectionOfListeners(){
		ModelWithMultipleListeners model = new ModelWithMultipleListeners();
		Listener1 assignedView = mockery.mock(Listener1.class);
		model.listenersListener1.add(assignedView);
		
		ModelViewAutoSetter<Model, Listener1> mvpAutoSetter 
				= new ModelViewAutoSetter<Model, Listener1>(model, assignedView);
		mvpAutoSetter.setViewAsModelListener();
		
		assertTrue(model.hasSingleInstanceOfListener1(assignedView));
	}
	
	@Test(expectedExceptions=MvpException.class)
	public void throwingIfListenerCollectionFieldIsNull(){
		ViewImplementingListener1AndListener2 view = new ViewImplementingListener1AndListener2();
		
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(new InvalidModelWithListeners(), view);
		autoSetter.setViewAsModelListener();
	}

	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	//TODO: move to another class
	//this test ensures that if View is not implementing any model listeners, then 
	//model must have at least one listener assigned. But this testing does not make any
	//sense in the currently presented method
	public void assertModelDoesHave_atLeastOneListenerInstance_OfAtLeastOneType_IfCustomListenerPresent(){
		ModelWithMultipleListeners model = new ModelWithMultipleListeners();
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(model, new Object());
		try{
			model.clearAllListeners();
			//initate with view that does not implement model's interface
			autoSetter.setViewAsModelListener();
			Assert.fail("exception must be thrown if model has no listener instances for " +
					"atleast one type of the listener");
		}
		catch(MvpException e){
			//ok
		}
		try{
			model.listeners = new HashSet<ListenerParent>();
			//initate with view that does not implement model's interface
			autoSetter.setViewAsModelListener();
			Assert.fail("exception must be thrown if model has no listener instances for " +
					"atleast one type of the listener");
		}
		catch(MvpException e){
			//ok
		}
		try{
			model.listeners = new HashSet<ListenerParent>();
			model.listeners.add(null);
			//initate with view that does not implement model's interface
			autoSetter.setViewAsModelListener();
			Assert.fail("exception must be thrown if model has no listener instances for " +
					"atleast one type of the listener");
		}
		catch(MvpException e){
			//ok
		}
		autoSetter = new ModelViewAutoSetter(model, new Object());
		model.clearAllListeners();
		model.listener2= mockery.mock(Listener2.class);
		autoSetter.setViewAsModelListener();
		//ok, no exceptions
		
		model.clearAllListeners();
		model.listeners = new HashSet<ListenerParent>();
		model.listeners.add(mockery.mock(ListenerParent.class, "init-model-listener-mock"));
		autoSetter.setViewAsModelListener();
		//ok, no exception
		
	}
	
	@Test(expectedExceptions=MvpException.class)
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	public void throwIfViewImplemnetsListenerInterface_ButModelHasNoFreeFieldForIt(){
		ModelWithSingleListener model = new ModelWithSingleListener();
		Listener1 viewImplementingModelListener = mockery.mock(Listener1.class, "view");
		Listener1 externalModeListener = new Mockery().mock(Listener1.class, "someListener");
		ModelViewAutoSetter autoSetter = new ModelViewAutoSetter(model, viewImplementingModelListener);
		model.listener = externalModeListener;
		
		autoSetter.setViewAsModelListener();
	}
	
	private static class ViewImplementingListener1AndListener2 implements Listener1, Listener2{
	}
	
	private static class InvalidModelWithListeners extends Model{
		@Listener(type=Listener1.class)
		private Collection<Listener1> nullCollection;
	}
	
}
