/*******************************************************************************
 * Copyright © 2011 Alexey Kaloshin. All Rights Reserved. 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 *  3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written permission. 
 * 
 *  THIS SOFTWARE IS PROVIDED BY Alexey Kaloshin "AS IS" AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
package cz.nocach.swingmvp.mvp;

import java.math.BigDecimal;
import java.util.HashSet;

import javax.swing.JTextField;

import org.jmock.Expectations;
import org.testng.annotations.Test;

import cz.nocach.swingmvp.BaseTest;
import cz.nocach.swingmvp.MvpException;
import cz.nocach.swingmvp.autoinitiater.MvpAutoInitiater;
import cz.nocach.swingmvp.autoinitiater.ViewAsModelSetter;
import cz.nocach.swingmvp.model.Model;
import cz.nocach.swingmvp.presenter.AutoSet;
import cz.nocach.swingmvp.presenter.BasePresenter;
import cz.nocach.swingmvp.presenter.Presenter;
import cz.nocach.swingmvp.testmodels.BaseModelDefaultChangeSupportViewStub;
import cz.nocach.swingmvp.testmodels.BaseModelWithCustomListeners;
import cz.nocach.swingmvp.testmodels.BaseTestModel;
import cz.nocach.swingmvp.testmodels.BaseTestModelListener;
import cz.nocach.swingmvp.testmodels.Product;
import cz.nocach.swingmvp.testobj.EmptyModel;
import cz.nocach.swingmvp.view.Change;

public class MvpInitiaterTest extends BaseTest{

	@Test
	public void autoSettingOfViewAsModelListener(){
		final ViewAsModelSetter viewAsModelSetter = mockery.mock(ViewAsModelSetter.class);
		MvpAutoInitiater mvpInitater = new MvpAutoInitiater(new EmptyModel(), new Object());
		mvpInitater.setModelViewAutoSetter(viewAsModelSetter);
		
		mockery.checking(new Expectations(){{
			one(viewAsModelSetter).setViewAsModelListener();
		}});
		
		mvpInitater.setViewAsModelListener();
		
		mockery.assertIsSatisfied();
	}
	
	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	public void autoinitOfViewFromModelUsingDefaultChangeSupport(){
		final BaseTestModel model = new BaseTestModel();
		populateInitModelFields(model);
		BaseModelDefaultChangeSupportViewStub view = new BaseModelDefaultChangeSupportViewStub();
		model.addPropertyChangeListener(view);
		MvpAutoInitiater autoInit = new MvpAutoInitiater(model, view);
		
		autoInit.initViewFromModel();

		assertViewHasSameValuesAsModel(model, view);
	}

	private void assertViewHasSameValuesAsModel(final BaseTestModel model
			, BaseModelDefaultChangeSupportViewStub view) {
		assertEquals(view.name, model.name);
		assertEquals(view.age, model.age);
		assertEquals(view.credit, model.credit);
		assertEquals(view.favoriteProduct, model.favoriteProduct);
		assertEquals(view.orderedProducts, model.orderedProducts);
	}
	
	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	public void autoinitOfViewFromModelUsingCustomListeners(){
		final BaseModelWithCustomListeners model = new BaseModelWithCustomListeners();
		populateInitModelFields(model);
		//as we are testing our view acting as listener of the model changes
		//we just mock BaseModelListener and pretend that view is something more then just a mock
		final BaseTestModelListener view = mockery.mock(BaseTestModelListener.class);
		model.listener = view;
		MvpAutoInitiater autoInit = new MvpAutoInitiater(model, view);
		//external listener should not be invoked in auto-init phase
		final BaseTestModelListener listenerThatShouldNotBeInvoked = mockery.mock(BaseTestModelListener.class, "anotherInitModelListener");
		model.listenerTwo = listenerThatShouldNotBeInvoked;
		
		mockery.checking(new Expectations(){{
			exactly(0).of(same(listenerThatShouldNotBeInvoked)).method("/*/");
			
			oneOf(view).onAgeChange(model.getAge());
			oneOf(view).onCreditChange(model.getCredit());
			oneOf(view).onFavoriteProductChange(model.getFavoriteProduct());
			oneOf(view).onNameChange(model.getName());
			
			for(Product each : model.orderedProducts){
				oneOf(view).onOrderAdded(each);
			}
		}});
		
		autoInit.initViewFromModel();
		
		mockery.assertIsSatisfied();
	}
	
	private void populateInitModelFields(final BaseTestModel model) {
		model.age = 77;
		model.credit = 42;
		model.favoriteProduct = new Product();
		model.favoriteProduct.setPrice(new BigDecimal(19999));
		model.favoriteProduct.setProductName("game");
		model.orderedProducts = new HashSet<Product>();
		
		Product orderProduct1 = new Product();
		orderProduct1.setPrice(new BigDecimal(1244));
		orderProduct1.setProductName("movie");
		model.orderedProducts.add(orderProduct1);
		model.orderedProducts.add(new Product());
	}
	
	@Test
	@SuppressWarnings({ "rawtypes", "unchecked" }) //Readability suffers otherwise
	public void autoViewPresenterEventDelegation(){
		BaseTestModel baseModel = new BaseTestModel();
		ViewForEventDelegating view = new ViewForEventDelegating();
		MvpAutoInitiater autoInit = new MvpAutoInitiater(baseModel, view);
		PresenterForEventDelegating presenter = new PresenterForEventDelegating(view, baseModel);
		
		autoInit.initViewPresenterAutoEventsDelegation(view, presenter);
		//TODO: this test is absolutely wrong. It must be checked using mocks
		//that WidgetBinder.bindTo(widget, presenter, methodOnPresenter)
		//is called
		String newTextToTest = "new text to change"; 
		view.nameTextField.setText(newTextToTest);
		
		assertTrue(presenter.onNameChangeCalled);
		assertEquals(presenter.onNameChangeNewNameValue, newTextToTest);
	}
	
	@Test
	public void creationgOfPresenterFromView_ThatHasNoPresenterTypeSpecified(){
		MvpAutoInitiater initiater = new MvpAutoInitiater(new EmptyModel(), new ViewWithoutPresenterAnnotation());
		BasePresenter<ViewWithoutPresenterAnnotation, EmptyModel>
		presenterForViewWithoutAnnotation = initiater.createPresenterForView();
		assertEquals(BasePresenter.class, presenterForViewWithoutAnnotation.getClass());
	}
	
	@Test
	public void autoCreatingPresenterSetAsAnnotationOnView(){
		MvpAutoInitiater initiater = new MvpAutoInitiater(new EmptyModel(), new ViewWithPresenterAnnotation());
		BasePresenter<ViewWithPresenterAnnotation, EmptyModel>
			presenterForViewWithAnnotation = initiater.createPresenterForView();
		assertEquals(presenterForViewWithAnnotation.getClass(), PresenterForViewWithAnnotation.class
				,"Created presenter didn't have class that was specified on View's");
	}
	
	@Test(expectedExceptions=MvpException.class)
	public void throwIfPresenterDoesNotHaveCorrectConstructorSignature(){
		MvpAutoInitiater initiater = new MvpAutoInitiater(new EmptyModel(), new ViewWithAnnotationForInvalidPresenter());
		initiater.createPresenterForView();
	}
	
	public static class ViewForEventDelegating {
		@Change("onNameChange")
		private JTextField nameTextField = new JTextField();
		public ViewForEventDelegating(){
		}
	}
	
	public static class PresenterForEventDelegating extends BasePresenter<ViewForEventDelegating, BaseTestModel>{
		private boolean onNameChangeCalled = false;
		private String onNameChangeNewNameValue = null;
		public PresenterForEventDelegating(ViewForEventDelegating view, BaseTestModel model) {
			super(view, model);
		}
		@AutoSet("name")
		public void onNameChange(String newName){
			onNameChangeCalled = true;
			onNameChangeNewNameValue = newName;
		}
	}
	
	private static class ViewWithoutPresenterAnnotation{
	}
	
	@Presenter(type=PresenterForViewWithAnnotation.class)
	private static class ViewWithPresenterAnnotation{
	}
	
	
	@Presenter(type=InvalidPresenter.class)
	private static class ViewWithAnnotationForInvalidPresenter{
	}
	
	public static class PresenterForViewWithAnnotation
			extends BasePresenter<ViewWithPresenterAnnotation, EmptyModel>{
		public PresenterForViewWithAnnotation(ViewWithPresenterAnnotation view,
				EmptyModel model) {
			super(view, model);
		}
	}
	
	public static class InvalidPresenter 
			extends BasePresenter<ViewWithAnnotationForInvalidPresenter, EmptyModel>{

		public InvalidPresenter(ViewWithAnnotationForInvalidPresenter view) {
			super(view, new EmptyModel());
		}
		
	}
	
}
